@@ -526,9 +526,15 @@ BOOST_PYTHON_MODULE(_driver)
     .value("SCHED_SPIN", CU_CTX_SCHED_SPIN)
     .value("SCHED_YIELD", CU_CTX_SCHED_YIELD)
     .value("SCHED_MASK", CU_CTX_SCHED_MASK)
-#if CUDAPP_CUDA_VERSION >= 2020
+#if CUDAPP_CUDA_VERSION >= 2020 && CUDAPP_CUDA_VERSION < 4000
+    .value("BLOCKING_SYNC", CU_CTX_BLOCKING_SYNC)
+    .value("SCHED_BLOCKING_SYNC", CU_CTX_BLOCKING_SYNC)
+#endif
+#if CUDAPP_CUDA_VERSION >= 4000
     .value("BLOCKING_SYNC", CU_CTX_SCHED_BLOCKING_SYNC)
     .value("SCHED_BLOCKING_SYNC", CU_CTX_SCHED_BLOCKING_SYNC)
+#endif
+#if CUDAPP_CUDA_VERSION >= 2020
     .value("MAP_HOST", CU_CTX_MAP_HOST)
 #endif
 #if CUDAPP_CUDA_VERSION >= 3020
